import sys
from base64 import b64decode, urlsafe_b64encode

import requests
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives.asymmetric.utils import encode_dss_signature
from cryptography.hazmat.primitives.serialization import load_der_public_key

from mdast_cli.helpers.logging import Log

from . import config, googleplay_pb2, utils

BASE = "https://android.clients.google.com/"
FDFE = BASE + "fdfe/"
CHECKIN_URL = BASE + "checkin"
AUTH_URL = BASE + "auth"

UPLOAD_URL = FDFE + "uploadDeviceConfig"
DETAILS_URL = FDFE + "details"
DELIVERY_URL = FDFE + "delivery"
PURCHASE_URL = FDFE + "purchase"

CONTENT_TYPE_URLENC = "application/x-www-form-urlencoded; charset=UTF-8"
CONTENT_TYPE_PROTO = "application/x-protobuf"


def encrypt_password(login, password):
    binaryKey = b64decode(config.GOOGLE_PUBKEY)
    i = utils.readInt(binaryKey, 0)
    modulus = utils.toBigInt(binaryKey[4:][0:i])
    j = utils.readInt(binaryKey, i + 4)
    exponent = utils.toBigInt(binaryKey[i + 8:][0:j])

    digest = hashes.Hash(hashes.SHA1(), backend=default_backend())
    digest.update(binaryKey)
    h = b'\x00' + digest.finalize()[0:4]

    der_data = encode_dss_signature(modulus, exponent)
    publicKey = load_der_public_key(der_data, backend=default_backend())

    to_be_encrypted = login.encode() + b'\x00' + password.encode()
    ciphertext = publicKey.encrypt(
        to_be_encrypted,
        padding.OAEP(
            mgf=padding.MGF1(algorithm=hashes.SHA1()),
            algorithm=hashes.SHA1(),
            label=None
        )
    )

    return urlsafe_b64encode(h + ciphertext)


class GooglePlayAPI(object):
    def __init__(self, locale="en_US", timezone="UTC", device_codename="angler"):
        self.authSubToken = None
        self.gsfId = None
        self.device_config_token = None
        self.deviceCheckinConsistencyToken = None
        self.dfeCookie = None
        self.deviceBuilder = config.DeviceBuilder(device_codename)
        self.deviceBuilder.setLocale(locale)
        self.deviceBuilder.setTimezone(timezone)

    def login(self, email, password, gsfId, authSubToken):
        if email is not None and password is not None:
            Log.info('Google Play - Logging in with email and password, you should copy token after')
            encryptedPass = encrypt_password(email, password).decode('utf-8')
            params = self.deviceBuilder.getLoginParams(email, encryptedPass)
            params['service'] = 'ac2dm'
            params['add_account'] = '1'
            params['callerPkg'] = 'com.google.android.gms'
            headers = self.deviceBuilder.getAuthHeaders(self.gsfId)
            headers['app'] = 'com.google.android.gsm'
            response = requests.post(AUTH_URL, data=params)
            data = response.text.split()
            params = {}
            for d in data:
                if "=" not in d:
                    continue
                k, v = d.split("=", 1)
                params[k.strip().lower()] = v.strip()
            if "auth" in params:
                ac2dmToken = params["auth"]
            elif "error" in params:
                if "NeedsBrowser" in params["error"]:
                    Log.error('Google Play - Security check is needed, '
                              'try to visit https://accounts.google.com/b/0/DisplayUnlockCaptcha to unlock.')
                    Log.error(f'Google Play - server says: "{params["error"]}"')
                    # sys.exit(4)

            self.gsfId = self.checkin(email, ac2dmToken)
            self.getAuthSubToken(email, encryptedPass)
            self.uploadDeviceConfig()
            Log.info(f'Google Play - gsfId: {self.gsfId}, authSubToken: {self.authSubToken}')
            Log.info(f'Google Play - You should copy these parameters and use them for next scans instead '
                     f'of email and password:')
            Log.info(f'Google Play - "--google_play_gsfid {self.gsfId} --google_play_auth_token {self.authSubToken}"')
        elif gsfId is not None and authSubToken is not None:
            self.gsfId = gsfId
            self.setAuthSubToken(authSubToken)
            Log.info('Google Play - Logging in with gsfid and auth token')
        else:
            Log.error('Google Play - Login failed.')
            sys.exit(4)

    def download(self, packageName, versionCode=None, offerType=1):
        if self.authSubToken is None:
            Log.error('Google Play - You need to login before executing any request')
            sys.exit(4)

        if versionCode is None:
            appDetails = self.details(packageName).get('details').get('appDetails')
            versionCode = appDetails.get('versionCode')

        headers = self.getHeaders()
        params = {'ot': str(offerType),
                  'doc': packageName,
                  'vc': str(versionCode)}
        response = requests.post(PURCHASE_URL, headers=headers,
                                 params=params,
                                 timeout=60)

        response = googleplay_pb2.ResponseWrapper.FromString(response.content)
        if response.commands.displayErrorMessage != "":
            Log.error(f'Google Play - {response.commands.displayErrorMessage}')
            sys.exit(4)
        else:
            downloadToken = response.payload.buyResponse.downloadToken

        if downloadToken is not None:
            params['dtok'] = downloadToken
        response = requests.get(DELIVERY_URL, headers=headers, params=params, timeout=60)
        response = googleplay_pb2.ResponseWrapper.FromString(response.content)
        if response.commands.displayErrorMessage != "":
            Log.error(f'Google Play - {response.commands.displayErrorMessage}')
            sys.exit(4)
        elif response.payload.deliveryResponse.appDeliveryData.downloadUrl == "":
            Log.error('Google Play - App not purchased')
            sys.exit(4)
        else:
            result = {'docId': packageName, 'additionalData': [], 'splits': []}
            downloadUrl = response.payload.deliveryResponse.appDeliveryData.downloadUrl
            cookie = response.payload.deliveryResponse.appDeliveryData.downloadAuthCookie[0]
            cookies = {
                str(cookie.name): str(cookie.value)
            }
            result['file'] = self._deliver_data(downloadUrl, cookies)

            for split in response.payload.deliveryResponse.appDeliveryData.split:
                a = {'name': split.name, 'file': self._deliver_data(split.downloadUrl, None)}
                result['splits'].append(a)
            return result, appDetails

    def setAuthSubToken(self, authSubToken):
        self.authSubToken = authSubToken

    def details(self, packageName):
        path = DETAILS_URL + f'?doc={requests.utils.quote(packageName)}'
        data = self.executeRequestApi2(path)
        return utils.parseProtobufObj(data.payload.detailsResponse.docV2)

    def getHeaders(self, upload_fields=False):
        if upload_fields:
            headers = self.deviceBuilder.getDeviceUploadHeaders()
        else:
            headers = self.deviceBuilder.getBaseHeaders()
        if self.gsfId is not None:
            headers["X-DFE-Device-Id"] = "{0:x}".format(self.gsfId)
        if self.authSubToken is not None:
            headers["Authorization"] = "GoogleLogin auth=%s" % self.authSubToken
        if self.device_config_token is not None:
            headers["X-DFE-Device-Config-Token"] = self.device_config_token
        if self.deviceCheckinConsistencyToken is not None:
            headers["X-DFE-Device-Checkin-Consistency-Token"] = self.deviceCheckinConsistencyToken
        if self.dfeCookie is not None:
            headers["X-DFE-Cookie"] = self.dfeCookie
        return headers

    def checkin(self, email, ac2dmToken):
        headers = self.getHeaders()
        headers["Content-Type"] = CONTENT_TYPE_PROTO

        request = self.deviceBuilder.getAndroidCheckinRequest()

        stringRequest = request.SerializeToString()
        res = requests.post(CHECKIN_URL, data=stringRequest,
                            headers=headers)
        response = googleplay_pb2.AndroidCheckinResponse()
        response.ParseFromString(res.content)
        self.deviceCheckinConsistencyToken = response.deviceCheckinConsistencyToken

        request.id = response.androidId
        request.securityToken = response.securityToken
        request.accountCookie.append("[" + email + "]")
        request.accountCookie.append(ac2dmToken)
        stringRequest = request.SerializeToString()
        requests.post(CHECKIN_URL,
                      data=stringRequest,
                      headers=headers)

        return response.androidId

    def uploadDeviceConfig(self):
        upload = googleplay_pb2.UploadDeviceConfigRequest()
        upload.deviceConfiguration.CopyFrom(self.deviceBuilder.getDeviceConfig())
        headers = self.getHeaders(upload_fields=True)
        stringRequest = upload.SerializeToString()
        response = requests.post(UPLOAD_URL, data=stringRequest,
                                 headers=headers,
                                 timeout=60)
        response = googleplay_pb2.ResponseWrapper.FromString(response.content)
        try:
            if response.payload.HasField('uploadDeviceConfigResponse'):
                self.device_config_token = response.payload.uploadDeviceConfigResponse
                self.device_config_token = self.device_config_token.uploadDeviceConfigToken
        except ValueError:
            pass

    def getAuthSubToken(self, email, passwd):
        requestParams = self.deviceBuilder.getLoginParams(email, passwd)
        requestParams['service'] = 'androidmarket'
        requestParams['app'] = 'com.android.vending'
        headers = self.deviceBuilder.getAuthHeaders(self.gsfId)
        headers['app'] = 'com.android.vending'
        response = requests.post(AUTH_URL,
                                 data=requestParams,
                                 headers=headers)
        data = response.text.split()
        params = {}
        for d in data:
            if "=" not in d:
                continue
            k, v = d.split("=", 1)
            params[k.strip().lower()] = v.strip()
        if "token" in params:
            master_token = params["token"]
            second_round_token = self.getSecondRoundToken(master_token, requestParams)
            self.authSubToken = second_round_token
        elif "error" in params:
            Log.error(f'Google Play - server says: " + {params["error"]}')
            sys.exit(4)
        else:
            Log.error('Google Play - auth token not found')
            sys.exit(4)

    def getSecondRoundToken(self, first_token, params):
        if self.gsfId is not None:
            params['androidId'] = "{0:x}".format(self.gsfId)
        params['Token'] = first_token
        params['check_email'] = '1'
        params['token_request_options'] = 'CAA4AQ=='
        params['system_partition'] = '1'
        params['_opt_is_called_from_account_manager'] = '1'
        params.pop('Email')
        params.pop('EncryptedPasswd')
        headers = self.deviceBuilder.getAuthHeaders(self.gsfId)
        headers['app'] = 'com.android.vending'
        response = requests.post(AUTH_URL,
                                 data=params,
                                 headers=headers)
        data = response.text.split()
        params = {}
        for d in data:
            if "=" not in d:
                continue
            k, v = d.split("=", 1)
            params[k.strip().lower()] = v.strip()

        if 'error' in params:
            raise ConnectionError('')

        if "auth" in params:
            return params["auth"]
        elif "error" in params:
            Log.error(f'Google Play - server says: " + {params["error"]}')
            sys.exit(4)
        else:
            Log.error('Google Play - auth token not found')
            sys.exit(4)

    def executeRequestApi2(self, path, post_data=None, content_type=CONTENT_TYPE_URLENC, params=None):
        if self.authSubToken is None:
            Log.error('You need to login before executing any request')
            sys.exit(4)
        headers = self.getHeaders()
        headers["Content-Type"] = content_type

        if post_data is not None:
            response = requests.post(path,
                                     data=str(post_data),
                                     headers=headers,
                                     params=params,
                                     timeout=60)
        else:
            response = requests.get(path,
                                    headers=headers,
                                    params=params,
                                    timeout=60)

        message = googleplay_pb2.ResponseWrapper.FromString(response.content)
        if message.commands.displayErrorMessage != "":
            raise ConnectionError(message.commands.displayErrorMessage)

        return message

    def _deliver_data(self, url, cookies):
        headers = self.getHeaders()
        response = requests.get(url, headers=headers, cookies=cookies, stream=True, timeout=60)
        total_size = response.headers.get('content-length')
        chunk_size = 32 * (1 << 10)
        return {'data': response.iter_content(chunk_size=chunk_size),
                'total_size': total_size,
                'chunk_size': chunk_size}



import os
import plistlib
import sys
import time
import zipfile

import requests
from requests.adapters import HTTPAdapter
from urllib3 import Retry

from .appstore_client.store import StoreClient, StoreException

try:
    from ..helpers.logging import Log
    from .base import DistributionSystem
except ImportError:
    from mdast_cli.distribution_systems.base import DistributionSystem
    from mdast_cli.helpers.logging import Log


def get_zipinfo_datetime(timestamp=None):
    timestamp = int(timestamp or time.time())
    return time.gmtime(timestamp)[0:6]


def download_file(url, download_path, file_path):
    with requests.get(url, stream=True) as r:
        if r.status_code != 200:
            Log.error(f'Failed to download application. Request return status code: {r.status_code}"')
            sys.exit(4)
        if not os.path.exists(download_path):
            os.mkdir(download_path)
            Log.info(f'Creating directory {download_path} for downloading app from AppStore')
        with open(file_path, 'wb') as f:
            for chunk in r.iter_content(chunk_size=1 * 1024 * 1024):
                f.write(chunk)
        f.close()
        if os.path.exists(file_path):
            Log.info('Application successfully downloaded')
        else:
            Log.info('Failed to download application. '
                     'Seems like something is wrong with your file path or app file is broken')

    return file_path


def _login_iTunes(Store, apple_id, pass2FA):
    Log.info('Logging into iTunes')
    Store.authenticate(apple_id, pass2FA)
    Log.info(f'Successfully logged in as {Store.account_name}')


class AppStore(DistributionSystem):
    """
    Downloading application(.ipa file) from AppStore
    """

    def __init__(self, appstore_apple_id, appstore_password2FA,
                 appstore_app_id=None, appstore_bundle_id=None, appstore_file_name=None):
        self.apple_id = appstore_apple_id
        self.pass2FA = appstore_password2FA
        self.app_id = appstore_app_id
        self.bundle_id = appstore_bundle_id
        self.appstore_file_name = appstore_file_name
        self.download_path = 'downloaded_apps'

        super().__init__(appstore_app_id, '')
        self.sess = requests.Session()

        retry_strategy = Retry(
            connect=4,
            read=2,
            total=8,
        )
        self.sess.mount("https://", HTTPAdapter(max_retries=retry_strategy))
        self.sess.mount("http://", HTTPAdapter(max_retries=retry_strategy))

    def download_app(self):
        try:
            Store = StoreClient(self.sess)
            _login_iTunes(Store, self.apple_id, self.pass2FA)
        except StoreException as e:
            Log.error("Store %s failed! Message: %s%s" % (
                e.req, e.err_msg, " (errorType %s)" % e.err_type if e.err_type else ''))
            Log.info('Failed to download application. Seems like your credentials are incorrect '
                     'or your 2FA code expired.')
            sys.exit(4)

        try:
            if self.app_id is None:
                Log.info(f'Trying to find app in App Store with bundle id {self.bundle_id}')
                found_by_bundle_resp = Store.find_app_by_bundle(self.bundle_id)
                resp_info = found_by_bundle_resp.json()
                if found_by_bundle_resp.status_code != 200 or resp_info['resultCount'] != 1:
                    Log.error('Application with your bundle id not found, probably you enter invalid bundle')
                    sys.exit(4)
                app_info = resp_info['results'][0]
                Log.info(f'Successfully found application by bundle id ({self.bundle_id}) '
                         f'with name: "{app_info["trackName"]}", version: {app_info["version"]},'
                         f' app_id: {app_info["trackId"]}')
                self.app_id = app_info["trackId"]

            Log.info(f'Trying to purchase app with id {self.app_id}')
            purchase_resp = Store.purchase(self.app_id)
            if purchase_resp.status_code == 200:
                Log.info(f'App was successfully purchased for {self.apple_id} account')
            elif purchase_resp.status_code == 500:
                Log.info(f'This app was purchased before for {self.apple_id} account')
            Log.info(f'Retrieving download info for app with id: {self.app_id}')
            download_resp = Store.download(self.app_id)
            if not download_resp.songList:
                Log.error('Failed to get app download info! Check your parameters')
                sys.exit(3)
            downloaded_app_info = download_resp.songList[0]

            app_name = downloaded_app_info.metadata.bundleDisplayName
            app_id = downloaded_app_info.songId
            app_bundle_id = downloaded_app_info.metadata.softwareVersionBundleId
            app_version = downloaded_app_info.metadata.bundleShortVersionString

            Log.info(
                f'Downloading app is {app_name} ({app_bundle_id}) with app_id {app_id} and version {app_version}')

            if self.appstore_file_name is None:
                file_name = '%s-%s.ipa' % (app_name, app_version)
            else:
                file_name = '%s-%s.ipa' % (self.appstore_file_name, app_version)

            file_path = os.path.join(self.download_path, file_name)
            Log.info(f'Downloading ipa to {file_path}')
            download_file(downloaded_app_info.URL, self.download_path, file_path)

            with zipfile.ZipFile(file_path, 'a') as ipa_file:
                Log.info('Creating iTunesMetadata.plist with metadata info')
                metadata = downloaded_app_info.metadata.as_dict()
                metadata["apple-id"] = self.apple_id
                metadata["userName"] = self.apple_id
                ipa_file.writestr(zipfile.ZipInfo("iTunesMetadata.plist", get_zipinfo_datetime()),
                                  plistlib.dumps(metadata))

                appContentDir = [c for c in ipa_file.namelist() if
                                 c.startswith('Payload/') and len(c.strip('/').split('/')) == 2][0]
                appContentDir = appContentDir.rstrip('/')

                scManifestData = ipa_file.read(appContentDir + '/SC_Info/Manifest.plist')
                scManifest = plistlib.loads(scManifestData)

                sinfs = {c.id: c.sinf for c in downloaded_app_info.sinfs}
                for i, sinfPath in enumerate(scManifest['SinfPaths']):
                    ipa_file.writestr(appContentDir + '/' + sinfPath, sinfs[i])

        except StoreException as e:
            Log.error("Store %s failed! Message: %s%s" % (
                e.req, e.err_msg, " (errorType %s)" % e.err_type if e.err_type else ''))
            Log.info('Failed to download application. Seems like your app_id does not exist or you did not purchase '
                     'this app from apple account before.')
            sys.exit(4)

        return file_path



import os
import sys

from .gpapi.googleplay import GooglePlayAPI

try:
    from ..helpers.logging import Log
except ImportError:
    from mdast_cli.helpers.logging import Log


def google_play_download(package_name,
                         email=None,
                         password=None,
                         gsfId=None,
                         authSubToken=None,
                         file_name=None,
                         download_with_creds=False):
    try:
        gp_api = GooglePlayAPI()

        Log.info('Google Play - Google Play integration, trying to login')
        gp_api.login(email, password, gsfId, authSubToken)

        if email is not None and password is not None and download_with_creds is False:
            sys.exit(0)
        else:
            download_path = 'downloaded_apps'

            downloaded_file, app_details = gp_api.download(package_name)

            if file_name is None:
                file_name = package_name

            app_version = app_details.get('versionString')
            path_to_file = f'{download_path}/{file_name}-v{app_version}.apk'
            Log.info('Google Play - Successfully logged in Play Store')
            Log.info(f'Google Play - Downloading {package_name} apk to {path_to_file}')

            if not os.path.exists(download_path):
                os.mkdir(download_path)
                Log.info(f'Google Play - Creating directory {download_path} for downloading app from Google Play Store')

            with open(path_to_file, 'wb') as file:
                for chunk in downloaded_file.get('file').get('data'):
                    file.write(chunk)

            if os.path.exists(path_to_file):
                Log.info('Google Play - Application successfully downloaded!')
            else:
                Log.info('Google Play - Failed to download application. '
                         'Seems like something is wrong with your file path or app file is broken')
                sys.exit(4)

        return path_to_file

    except Exception as e:
        Log.error(f'Google Play - Failed to download application. Seems like something goes wrong.'
                  f' {e}')
        Log.info('Google Play - Try to reinitialize your account by logging in with your password and email')
        sys.exit(4)



from reprlib import repr as limitedRepr


class StoreDownloadReq:

    _types_map = {
        "creditDisplay": {"type": str, "subtype": None},
        "guid": {"type": str, "subtype": None},
        "salableAdamId": {"type": str, "subtype": None},
        "appExtVrsId": {"type": str, "subtype": None},
    }
    _formats_map = {}
    _validations_map = {
        "creditDisplay": {
            "required": True,
        },
        "guid": {
            "required": True,
        },
        "salableAdamId": {
            "required": True,
        },
        "appExtVrsId": {
            "required": False,
        },
    }

    def __init__(
        self,
        creditDisplay: str = None,
        guid: str = None,
        salableAdamId: str = None,
        appExtVrsId: str = None,
    ):
        pass
        self.__creditDisplay = creditDisplay
        self.__guid = guid
        self.__salableAdamId = salableAdamId
        self.__appExtVrsId = appExtVrsId

    def _get_creditDisplay(self):
        return self.__creditDisplay

    def _set_creditDisplay(self, value):
        if not isinstance(value, str):
            raise TypeError("creditDisplay must be str")

        self.__creditDisplay = value

    creditDisplay = property(_get_creditDisplay, _set_creditDisplay)

    def _get_guid(self):
        return self.__guid

    def _set_guid(self, value):
        if not isinstance(value, str):
            raise TypeError("guid must be str")

        self.__guid = value

    guid = property(_get_guid, _set_guid)

    def _get_salableAdamId(self):
        return self.__salableAdamId

    def _set_salableAdamId(self, value):
        if not isinstance(value, str):
            raise TypeError("salableAdamId must be str")

        self.__salableAdamId = value

    salableAdamId = property(_get_salableAdamId, _set_salableAdamId)

    def _get_appExtVrsId(self):
        return self.__appExtVrsId

    def _set_appExtVrsId(self, value):
        if value is not None and not isinstance(value, str):
            raise TypeError("appExtVrsId must be str")

        self.__appExtVrsId = value

    appExtVrsId = property(_get_appExtVrsId, _set_appExtVrsId)

    @staticmethod
    def from_dict(d):
        v = {}
        if "creditDisplay" in d:
            v["creditDisplay"] = (
                str.from_dict(d["creditDisplay"])
                if hasattr(str, "from_dict")
                else d["creditDisplay"]
            )
        if "guid" in d:
            v["guid"] = (
                str.from_dict(d["guid"]) if hasattr(str, "from_dict") else d["guid"]
            )
        if "salableAdamId" in d:
            v["salableAdamId"] = (
                str.from_dict(d["salableAdamId"])
                if hasattr(str, "from_dict")
                else d["salableAdamId"]
            )
        if "appExtVrsId" in d:
            v["appExtVrsId"] = (
                str.from_dict(d["appExtVrsId"])
                if hasattr(str, "from_dict")
                else d["appExtVrsId"]
            )
        return StoreDownloadReq(**v)

    def as_dict(self):
        d = {}
        if self.__creditDisplay is not None:
            d["creditDisplay"] = (
                self.__creditDisplay.as_dict()
                if hasattr(self.__creditDisplay, "as_dict")
                else self.__creditDisplay
            )
        if self.__guid is not None:
            d["guid"] = (
                self.__guid.as_dict()
                if hasattr(self.__guid, "as_dict")
                else self.__guid
            )
        if self.__salableAdamId is not None:
            d["salableAdamId"] = (
                self.__salableAdamId.as_dict()
                if hasattr(self.__salableAdamId, "as_dict")
                else self.__salableAdamId
            )
        if self.__appExtVrsId is not None:
            d["appExtVrsId"] = (
                self.__appExtVrsId.as_dict()
                if hasattr(self.__appExtVrsId, "as_dict")
                else self.__appExtVrsId
            )
        return d

    def __repr__(self):
        return "<Class StoreDownloadReq. creditDisplay: {}, guid: {}, salableAdamId: {}, appExtVrsId: {}>".format(
            limitedRepr(
                self.__creditDisplay[:20]
                if isinstance(self.__creditDisplay, bytes)
                else self.__creditDisplay
            ),
            limitedRepr(
                self.__guid[:20] if isinstance(self.__guid, bytes) else self.__guid
            ),
            limitedRepr(
                self.__salableAdamId[:20]
                if isinstance(self.__salableAdamId, bytes)
                else self.__salableAdamId
            ),
            limitedRepr(
                self.__appExtVrsId[:20]
                if isinstance(self.__appExtVrsId, bytes)
                else self.__appExtVrsId
            ),
        )



import argparse
import json
import os
import sys
import time

import urllib3

try:
    from distribution_systems.appstore import AppStore
    from distribution_systems.google_play import google_play_download
    from helpers.const import END_SCAN_TIMEOUT, SLEEP_TIMEOUT, TRY_COUNT, DastState, DastStateDict
    from helpers.logging import Log
except ImportError:
    from mdast_cli.distribution_systems.appstore import AppStore
    from mdast_cli.distribution_systems.google_play import google_play_download
    from mdast_cli.helpers.const import END_SCAN_TIMEOUT, SLEEP_TIMEOUT, TRY_COUNT, DastState, DastStateDict
    from mdast_cli.helpers.logging import Log


def parse_args():
    parser = argparse.ArgumentParser(description='Start download and get file.')
    parser.add_argument('--distribution_system', '-ds', type=str, help='Select how to download file: '
                                                                       'appstore/google_play',
                        choices=['appstore', 'google_play'], required=True)
    # Arguments for AppStore
    parser.add_argument('--appstore_app_id', type=str,
                        help='Application id from AppStore, you can get it on app page from url, format: .../id{APP_ID}'
                             'You should specify either bundle_id or app_id if distribution system set to "appstore"')
    parser.add_argument('--appstore_bundle_id', type=str,
                        help='Bundle id of application id from AppStore.'
                             ' You should specify either bundle_id or app_id if distribution system set to "appstore"')
    parser.add_argument('--appstore_apple_id', type=str,
                        help='Your email for iTunes login. This argument is required '
                             'if distribution system set to "appstore"')
    parser.add_argument('--appstore_password2FA', type=str,
                        help='Your password and 2FA code for iTunes login, format: password2FA_code (password1337)'
                             'This argument is required if distribution system set to "appstore"')
    parser.add_argument('--appstore_file_name', type=str,
                        help='File name for downloaded application.'
                             ' This argument is optional if distribution system set to "appstore"')

    # Arguments for Google Play
    parser.add_argument('--google_play_package_name', type=str,
                        help='Application package name from Google Play. This argument is required for first login if '
                             'distribution system set to "google_play"')
    parser.add_argument('--google_play_email', type=str,
                        help='Your email for Google Play login. This argument is required for first login if '
                             'distribution system set to "google_play"')
    parser.add_argument('--google_play_password', type=str,
                        help='Your password for Google Play login. This argument is required for first login if'
                             'distribution system set to "google_play"')
    parser.add_argument('--google_play_gsfid', type=int,
                        help='The Google Services Framework Identifier (GSF ID) is a unique 16 character hexadecimal '
                             'number of your fake device. You should get it from console logs during '
                             'first login attempt with email and password.This argument required if '
                             'distribution system set to "google_play"')
    parser.add_argument('--google_play_auth_token', type=str,
                        help='Google auth token for access to Google Play API. '
                             'You should get it from console logs during first login attempt with email and password.'
                             ' This argument required if distribution system set to "google_play"')
    parser.add_argument('--google_play_file_name', type=str,
                        help='File name for downloaded application.'
                             ' This argument is optional if distribution system set to "google_play"')
    parser.add_argument('--google_play_download_with_creds', action='store_true',
                        help='Download the application at the first login with email + password. '
                             'This argument is optional if distribution system set to "google_play"')

    args = parser.parse_args()

    if args.distribution_system == 'appstore' and (
            (args.appstore_app_id is None and args.appstore_bundle_id is None) or
            args.appstore_apple_id is None or
            args.appstore_password2FA is None):
        parser.error('"--distribution_system appstore" requires either "--appstore_app_id" or "--appstore_bundle_id", '
                     '"--appstore_apple_id", "--appstore_password2FA" arguments to be set')

    elif args.distribution_system == 'google_play' and (
            args.google_play_package_name is None or
            ((args.google_play_email is None or args.google_play_password is None) and
             (args.google_play_gsfid is None or args.google_play_auth_token is None))):
        parser.error('"--distribution_system google_play" requires "--google_play_package_name" and either'
                     ' email + pass ("--google_play_email" + "--google_play_password") or '
                     'gsfid + token ("--google_play_gsfid" + "google_play_auth_token") arguments to be set')
    return args


def main():
    urllib3.disable_warnings()

    arguments = parse_args()

    distribution_system = arguments.distribution_system
    app_file = ''
    if distribution_system == 'appstore':
        appstore = AppStore(arguments.appstore_apple_id,
                            arguments.appstore_password2FA,
                            arguments.appstore_app_id,
                            arguments.appstore_bundle_id,
                            arguments.appstore_file_name)
        app_file = appstore.download_app()
    elif distribution_system == 'google_play':
        app_file = google_play_download(arguments.google_play_package_name,
                                        arguments.google_play_email,
                                        arguments.google_play_password,
                                        arguments.google_play_gsfid,
                                        arguments.google_play_auth_token,
                                        arguments.google_play_file_name,
                                        arguments.google_play_download_with_creds)

    if arguments.download_only is True:
        Log.info('Your application was downloaded!')
        sys.exit(0)


if __name__ == '__main__':
    main()



class DastState:
    CREATED = 0
    STARTING = 1
    STARTED = 2
    ANALYZING = 3
    SUCCESS = 4
    FAILED = 5
    STOPPING = 6
    RECALCULATING = 7
    INTERRUPTING = 8
    INITIALIZING = 9
    CANCELLED = 10
    CANCELLING = 11


DastStateDict = {
    0: "CREATED",
    1: "STARTING",
    2: "STARTED",
    3: "ANALYZING",
    4: "SUCCESS",
    5: "FAILED",
    6: "STOPPING",
    7: "RECALCULATING",
    8: "INTERRUPTING",
    9: "INITIALIZING",
    10: "CANCELLED",
    11: "CANCELLING"
}


TRY_COUNT = 360
END_SCAN_TIMEOUT = 30
SLEEP_TIMEOUT = 10



import plistlib

import requests

from .schemas.store_authenticate_req import StoreAuthenticateReq
from .schemas.store_authenticate_resp import StoreAuthenticateResp
from .schemas.store_download_req import StoreDownloadReq
from .schemas.store_download_resp import StoreDownloadResp
from .schemas.store_buyproduct_req import StoreBuyproductReq
from .schemas.store_buyproduct_resp import StoreBuyproductResp
from mdast_cli.helpers.logging import Log


class StoreException(Exception):
    def __init__(self, req, err_msg, err_type=None):
        self.req = req
        self.err_msg = err_msg
        self.err_type = err_type
        super().__init__(
            "Store %s error: %s" % (self.req, self.err_msg) if not self.err_type else
            "Store %s error: %s, errorType: %s" % (self.req, self.err_msg, self.err_type)
        )


class StoreClient(object):
    def __init__(self, sess: requests.Session, guid: str = '000C2941396B'):
        self.sess = sess
        self.guid = guid
        self.dsid = None
        self.store_front = None
        self.account_name = None

    def authenticate(self, appleId, password):
        req = StoreAuthenticateReq(appleId=appleId, password=password, attempt='4', createSession="true",
                                   guid=self.guid, rmp='0', why='signIn')
        url = "https://p46-buy.itunes.apple.com/WebObjects/MZFinance.woa/wa/authenticate?guid=%s" % self.guid
        while True:
            r = self.sess.post(url,
                               headers={
                                   "Accept": "*/*",
                                   "Content-Type": "application/x-www-form-urlencoded",
                                   "User-Agent":
                                       "Configurator/2.0 (Macintosh; OS X 10.12.6; 16G29) AppleWebKit/2603.3.8",
                               }, data=plistlib.dumps(req.as_dict()), allow_redirects=False)
            if r.status_code == 302:
                url = r.headers['Location']
                continue
            break
        resp = StoreAuthenticateResp.from_dict(plistlib.loads(r.content))
        if not resp.m_allowed:
            raise StoreException("authenticate", resp.customerMessage, resp.failureType)

        self.sess.headers['X-Dsid'] = self.sess.headers['iCloud-Dsid'] = str(resp.download_queue_info.dsid)
        self.sess.headers['X-Apple-Store-Front'] = r.headers.get('x-set-apple-store-front')
        self.sess.headers['X-Token'] = resp.passwordToken

        self.account_name = resp.accountInfo.address.firstName + " " + resp.accountInfo.address.lastName
        return resp

    def find_app_by_bundle(self, bundleId, appId=None, term=None, country="US", limit=1, media="software"):
        return self.sess.get("https://itunes.apple.com/lookup?",
                             params={
                                 "bundleId": bundleId,
                                 "id": appId,
                                 "term": term,
                                 "country": country,
                                 "limit": limit,
                                 "media": media,
                             },
                             headers={
                                 "Content-Type": "application/x-www-form-urlencoded",
                             })

    def purchase(self, appId, productType='C'):
        url = "https://buy.itunes.apple.com/WebObjects/MZBuy.woa/wa/buyProduct"
        req = StoreBuyproductReq(
            guid=self.guid,
            salableAdamId=str(appId),
            appExtVrsId='0',

            price='0',
            productType=productType,
            pricingParameters='STDQ',

            hasAskedToFulfillPreorder='true',
            buyWithoutAuthorization='true',
            hasDoneAgeCheck='true',
        )
        payload = req.as_dict()

        return self.sess.post(url,
                              headers={
                                  "Content-Type": "application/x-apple-plist",
                                  "User-Agent": "Configurator/2.15 (Macintosh; OS X 11.0.0; 16G29) AppleWebKit/2603.3.8",
                              },
                              data=plistlib.dumps(payload))

    def download(self, app_id, app_ver_id=""):
        req = StoreDownloadReq(creditDisplay="", guid=self.guid, salableAdamId=app_id, appExtVrsId=app_ver_id)
        r = self.sess.post("https://p25-buy.itunes.apple.com/WebObjects/MZFinance.woa/wa/volumeStoreDownloadProduct",
                           params={
                               "guid": self.guid
                           },
                           headers={
                               "Content-Type": "application/x-www-form-urlencoded",
                               "User-Agent": "Configurator/2.0 (Macintosh; OS X 10.12.6; 16G29) AppleWebKit/2603.3.8",
                           }, data=plistlib.dumps(req.as_dict()))

        resp = StoreDownloadResp.from_dict(plistlib.loads(r.content))
        if resp.cancel_purchase_batch:
            raise StoreException("volumeStoreDownloadProduct", resp, resp.customerMessage,
                                 resp.failureType + '-' + resp.metrics.dialogId)
        return resp



from reprlib import repr as limitedRepr


class StoreAuthenticateReq:

    _types_map = {
        "appleId": {"type": str, "subtype": None},
        "attempt": {"type": str, "subtype": None},
        "createSession": {"type": str, "subtype": None},
        "guid": {"type": str, "subtype": None},
        "password": {"type": str, "subtype": None},
        "rmp": {"type": str, "subtype": None},
        "why": {"type": str, "subtype": None},
    }
    _formats_map = {}
    _validations_map = {
        "appleId": {
            "required": True,
        },
        "attempt": {
            "required": True,
        },
        "createSession": {
            "required": True,
        },
        "guid": {
            "required": True,
        },
        "password": {
            "required": True,
        },
        "rmp": {
            "required": True,
        },
        "why": {
            "required": True,
        },
    }

    def __init__(
        self,
        appleId: str = None,
        attempt: str = None,
        createSession: str = None,
        guid: str = None,
        password: str = None,
        rmp: str = None,
        why: str = None,
    ):
        pass
        self.__appleId = appleId
        self.__attempt = attempt
        self.__createSession = createSession
        self.__guid = guid
        self.__password = password
        self.__rmp = rmp
        self.__why = why

    def _get_appleId(self):
        return self.__appleId

    def _set_appleId(self, value):
        if not isinstance(value, str):
            raise TypeError("appleId must be str")

        self.__appleId = value

    appleId = property(_get_appleId, _set_appleId)

    def _get_attempt(self):
        return self.__attempt

    def _set_attempt(self, value):
        if not isinstance(value, str):
            raise TypeError("attempt must be str")

        self.__attempt = value

    attempt = property(_get_attempt, _set_attempt)

    def _get_createSession(self):
        return self.__createSession

    def _set_createSession(self, value):
        if not isinstance(value, str):
            raise TypeError("createSession must be str")

        self.__createSession = value

    createSession = property(_get_createSession, _set_createSession)

    def _get_guid(self):
        return self.__guid

    def _set_guid(self, value):
        if not isinstance(value, str):
            raise TypeError("guid must be str")

        self.__guid = value

    guid = property(_get_guid, _set_guid)

    def _get_password(self):
        return self.__password

    def _set_password(self, value):
        if not isinstance(value, str):
            raise TypeError("password must be str")

        self.__password = value

    password = property(_get_password, _set_password)

    def _get_rmp(self):
        return self.__rmp

    def _set_rmp(self, value):
        if not isinstance(value, str):
            raise TypeError("rmp must be str")

        self.__rmp = value

    rmp = property(_get_rmp, _set_rmp)

    def _get_why(self):
        return self.__why

    def _set_why(self, value):
        if not isinstance(value, str):
            raise TypeError("why must be str")

        self.__why = value

    why = property(_get_why, _set_why)

    @staticmethod
    def from_dict(d):
        v = {}
        if "appleId" in d:
            v["appleId"] = (
                str.from_dict(d["appleId"])
                if hasattr(str, "from_dict")
                else d["appleId"]
            )
        if "attempt" in d:
            v["attempt"] = (
                str.from_dict(d["attempt"])
                if hasattr(str, "from_dict")
                else d["attempt"]
            )
        if "createSession" in d:
            v["createSession"] = (
                str.from_dict(d["createSession"])
                if hasattr(str, "from_dict")
                else d["createSession"]
            )
        if "guid" in d:
            v["guid"] = (
                str.from_dict(d["guid"]) if hasattr(str, "from_dict") else d["guid"]
            )
        if "password" in d:
            v["password"] = (
                str.from_dict(d["password"])
                if hasattr(str, "from_dict")
                else d["password"]
            )
        if "rmp" in d:
            v["rmp"] = (
                str.from_dict(d["rmp"]) if hasattr(str, "from_dict") else d["rmp"]
            )
        if "why" in d:
            v["why"] = (
                str.from_dict(d["why"]) if hasattr(str, "from_dict") else d["why"]
            )
        return StoreAuthenticateReq(**v)

    def as_dict(self):
        d = {}
        if self.__appleId is not None:
            d["appleId"] = (
                self.__appleId.as_dict()
                if hasattr(self.__appleId, "as_dict")
                else self.__appleId
            )
        if self.__attempt is not None:
            d["attempt"] = (
                self.__attempt.as_dict()
                if hasattr(self.__attempt, "as_dict")
                else self.__attempt
            )
        if self.__createSession is not None:
            d["createSession"] = (
                self.__createSession.as_dict()
                if hasattr(self.__createSession, "as_dict")
                else self.__createSession
            )
        if self.__guid is not None:
            d["guid"] = (
                self.__guid.as_dict()
                if hasattr(self.__guid, "as_dict")
                else self.__guid
            )
        if self.__password is not None:
            d["password"] = (
                self.__password.as_dict()
                if hasattr(self.__password, "as_dict")
                else self.__password
            )
        if self.__rmp is not None:
            d["rmp"] = (
                self.__rmp.as_dict() if hasattr(self.__rmp, "as_dict") else self.__rmp
            )
        if self.__why is not None:
            d["why"] = (
                self.__why.as_dict() if hasattr(self.__why, "as_dict") else self.__why
            )
        return d

    def __repr__(self):
        return "<Class StoreAuthenticateReq. appleId: {}, attempt: {}, createSession: {}, guid: {}, password: {}, rmp: {}, why: {}>".format(
            limitedRepr(
                self.__appleId[:20]
                if isinstance(self.__appleId, bytes)
                else self.__appleId
            ),
            limitedRepr(
                self.__attempt[:20]
                if isinstance(self.__attempt, bytes)
                else self.__attempt
            ),
            limitedRepr(
                self.__createSession[:20]
                if isinstance(self.__createSession, bytes)
                else self.__createSession
            ),
            limitedRepr(
                self.__guid[:20] if isinstance(self.__guid, bytes) else self.__guid
            ),
            limitedRepr(
                self.__password[:20]
                if isinstance(self.__password, bytes)
                else self.__password
            ),
            limitedRepr(
                self.__rmp[:20] if isinstance(self.__rmp, bytes) else self.__rmp
            ),
            limitedRepr(
                self.__why[:20] if isinstance(self.__why, bytes) else self.__why
            ),
        )
